import os
import warnings
from socket import AF_INET, SOCK_STREAM, socket
from typing import Union

import pyngrok.conf
import pyngrok.ngrok
import requests
import yaml
from pyngrok.exception import PyngrokError
from requests.exceptions import ConnectionError, InvalidURL

from . import env, models

logger = models.ngrok_logger()


class Ngrok:
    """Class to initiate ngrok connection, tunneling, and get existing tunnels.

    >>> Ngrok

    """

    SOCKET = socket(AF_INET, SOCK_STREAM)
    TUNNELS = f'http://{env.host}:4040/api/tunnels'
    CONNECTION = None

    def get_ngrok(self) -> Union[str, None]:
        """Identifies if an existing ngrok tunnel by sending a ``GET`` request to ngrok API endpoint.

        Returns:
            str:
            Returns the `ngrok` public URL.
        """
        try:
            logger.info(f'Looking for existing tunnels at {self.TUNNELS}')
            response = requests.get(url=self.TUNNELS)
        except InvalidURL:
            logger.error(f'Invalid URL: {self.TUNNELS}')
            return
        except ConnectionError:
            logger.error(f'Connection failed: {self.TUNNELS}')
            return
        if not response.ok:
            logger.error(f'Failed response [{response.status_code}] from {self.TUNNELS}')
            return
        serving_at = yaml.load(response.content.decode(), Loader=yaml.FullLoader)['tunnels']
        return serving_at[0].get('public_url')

    def connect(self, new_connection: bool = False) -> Union[str, None]:
        """Creates an HTTP socket and uses `pyngrok` module to bind the socket.

        Args:
            new_connection: Takes a boolean flag to bind a local socket to the port and spin up a new connection.

        See Also:
            Run the following code to setup.

            .. code-block:: python
                :emphasize-lines: 4,7,10

                from pyngrok.conf import PyngrokConfig, get_default
                from pyngrok.ngrok import set_auth_token

                # Sets auth token only during run time without modifying global config.
                PyngrokConfig.auth_token = '<NGROK_AUTH_TOKEN>'

                # Uses auth token from the specified file without modifying global config.
                get_default().config_path = "/path/to/config.yml"

                # Changes auth token at $HOME/.ngrok2/ngrok.yml
                set_auth_token('<NGROK_AUTH_TOKEN>')

        Returns:
            tuple:
            A tuple of the connected socket and the public URL.
        """
        if new_connection:
            server_address = (env.host, env.port)
            self.SOCKET.bind(server_address)  # Bind only accepts tuples
        self.SOCKET.listen(1)

        if env.ngrok_auth:
            logger.info('Using env var to set ngrok auth.')
            pyngrok.ngrok.set_auth_token(env.ngrok_auth)
        elif os.path.isfile('ngrok.yml'):
            logger.info('Using config file for ngrok auth')
            pyngrok.conf.get_default().config_path = 'ngrok.yml'
        else:
            logger.warning('Neither config file nor env var for ngrok auth was found.')
            return
        try:
            endpoint = pyngrok.ngrok.connect(env.port, "http", options={"remote_addr": f"{env.host}:{env.port}"})
            endpoint = endpoint.public_url.replace('http', 'https')
            logger.info(f'Ngrok connection has been established at {endpoint}.')
            return endpoint
        except PyngrokError as err:
            logger.error(err)

    def tunnel(self) -> None:
        """Once the socket is bound, the listener is activated and runs in a forever loop accepting connections."""
        while True:
            if env.STOPPER:
                break
            try:
                logger.info("Waiting for a connection")
                self.CONNECTION, client_address = self.SOCKET.accept()
                logger.info(f"Connection established from {client_address}")
            except KeyboardInterrupt:
                self.stop_tunnel()
                break
            except OSError:
                warnings.warn(
                    'Tunnel was initiated but the connection was never awaited.\n'
                    'If triggered in a thread, run as daemon or include a sleep time.'
                )
                break

    def stop_tunnel(self) -> None:
        """Closes the existing socket and connection where the socket is accepted."""
        logger.info("Shutting down server")
        if self.CONNECTION:
            self.CONNECTION.close()
        pyngrok.ngrok.kill(pyngrok_config=None)  # uses default config when None is passed
        self.SOCKET.close()



import base64
import binascii
import os
from datetime import datetime
from http.server import SimpleHTTPRequestHandler
from pathlib import PurePath
from time import time
from urllib.request import urlopen

import yaml
from gmailconnector.send_email import SendEmail

from . import env, models, ngrok

logger = models.server_logger()

HTML_PATH = f"{PurePath(__file__).parent}{os.path.sep}html{os.path.sep}"

with open(f'{HTML_PATH}auth_server.html') as file:
    auth_success = file.read()

with open(f'{HTML_PATH}no_auth.html') as file:
    login_failed = file.read()

with open(f'{HTML_PATH}session.html') as file:
    session_expiry = file.read()


# noinspection PyPep8Naming
class Authenticator(SimpleHTTPRequestHandler):
    """Main class to present webpages and authentication.

    >>> Authenticator

    Notes:
        - `Authenticator` uses encrypted server side header authentication.
        - Browser's default pop up will be shown prompting the user to enter the username and password.
        - Enter the Username and Password that was set in `environment variables <https://git.io/JCfzq>`__
        - The username and password are set as encoded auth headers, which are matched against the encoded env vars.
        - Upon successful authentication, a welcome page loads. Click on proceed to access the FileServer.

    See Also:
        - To perform authentication, a `do_AUTH()` is implemented sending 401 and WWW-Authenticate header in response.
        - To handle `POST` requests, a `do_POST()` has been implemented, which reads `content-length` header.
        - Headers are read using `rfile.read` from the start of the optional input data.
        - Data is sent to server using, `wfile.write` which contains the output stream to write response to the client.
    """

    def __init__(self, *args: tuple, **kwargs: dict):
        """Gets the authentication details from the user and encodes it.

        Args:
            *args: Socket generated using IP address and Port.
            **kwargs: Dictionary with key-value pairs of username, password and directory to serve.
        """
        self._auth = base64.b64encode(f"{kwargs.pop('username')}:{kwargs.pop('password')}".encode()).decode()
        super().__init__(*args, **kwargs)

    def log_message(self, format_: str, *args: tuple) -> None:
        """Suppresses logs from http.server by holding the args.

        Overwrites the base method: `BaseHTTPRequestHandler.log_message()`

        Args:
            format_: String operator %s
            *args: Logs from SimpleHTTPRequestHandler displaying request method type and HTTP status code.
        """
        if len(args) == 3:  # We don't have to care about NOT FOUND items because they are mostly hidden/protected.
            method, status_code, ignore = args  # ignore always returns `-`
            method = str(method).split('/')[0].strip()
            if int(str(status_code).strip()) == 200:
                logger.info(f'Received {status_code} while accessing a {method} method to reach {self.path}')
            else:
                logger.error(f'Received {status_code} while accessing a {method} method to reach {self.path}')

    def do_HEAD(self) -> None:
        """Sends 200 response and sends headers when authentication is successful."""
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()

    def do_GET(self) -> None:
        """Serve a front end with user authentication."""
        if _reset_auth() and 'Authorization' in self.headers.keys():
            logger.warning('Authorized via stored cookies. However session expired, so headers have been reset.')
            self.headers.replace_header('Authorization', None)
            self.disable_cache()

        if not (auth_header := self.headers.get("Authorization")):
            self.do_AUTH()
            if not env.authenticated:
                self.wfile.write(login_failed.encode(encoding='UTF-8'))
            else:
                self.wfile.write(session_expiry.encode(encoding='UTF-8'))
        elif auth_header == "Basic " + self._auth:
            target_path = env.host_dir + self.path
            if os.path.isdir(target_path) and 'index.html' in os.listdir(target_path):
                old_name = target_path + 'index.html'
                new_name = target_path + 'index_TEMP.html'
                logger.critical(f'Renaming {old_name} to {new_name}')
                os.rename(old_name, new_name)
                env.renamed.append({old_name: new_name})
            if not env.authenticated:
                self.do_HEAD()
                self.wfile.write(auth_success.encode(encoding='UTF-8'))
                env.authenticated = True
                return
            try:
                SimpleHTTPRequestHandler.do_GET(self)
            except BrokenPipeError:
                logger.error(f'Received BrokenPipeError while reaching {self.path}')
        else:
            self.do_AUTH()
            auth = auth_header.strip('Basic ')
            try:
                auth = base64.b64decode(auth).decode().split(':')
                logger.error(f'Authentication Blocked: Username: {auth[0]}\tPassword: {auth[1]}')
            except binascii.Error:
                logger.error(f'Authentication Blocked: Encoded: {auth}')

    def do_AUTH(self) -> None:
        """Sends headers to authenticate the requester."""
        self.send_response(401)
        self.send_header("WWW-Authenticate", 'Basic realm="Test"')
        self.send_header("Content-type", "text/html")
        self.end_headers()

    def do_POST(self) -> None:
        """Handles POST request and writes the received data into a yaml file."""
        now = datetime.now()
        client_file = 'client_info.yaml'
        length = int(self.headers.get('content-length'))
        peanut = self.rfile.read(length)
        butter = peanut.decode(encoding='UTF-8').split('&')[0]
        if butter == 'LOGOUT':
            logger.critical("Logout has been requested. Resetting auth headers to False.")
            env.first_run = True  # to reset and give a fresh start
            env.authenticated = False  # to show Login Success screen the next time
            return
        client_info = yaml.load(butter, Loader=yaml.FullLoader)
        if client_info.get('ip') == yaml.load(urlopen('https://ipapi.co/json/'), Loader=yaml.FullLoader).get('ip'):
            logger.info(f"Internal connection request received. Response: {client_info.get('_html_ref')}")
            return
        logger.fatal(str(client_info).strip("{}").replace("'", ""))

        if os.path.isfile(client_file):
            with open(client_file, 'r') as client:
                exist = yaml.load(''.join([nut for nut in client.readlines() if not nut.startswith('#')]),
                                  Loader=yaml.FullLoader)

            # avoids duplicate notification for repeated trials in less than 5 minutes
            if int(now.timestamp()) - int(os.stat(client_file).st_mtime) < 300 and \
                    client_info.get('_html_ref') == exist.get('_html_ref') and \
                    client_info.get('ip') == exist.get('ip'):
                logger.critical(f"{exist.get('ip')} performed {exist.get('_html_ref')} once again within 5 minutes")
                return

        current_time = f"Server Datetime: {now.strftime('%B %d, %Y %I:%M %p')}"
        hashes = ''.join(['#' for _ in range(74)])
        with open(client_file, 'w') as client:
            client.write(f"{hashes}\n#\t\t\t\t{current_time}\n{hashes}\n")
            yaml.dump(client_info, client, indent=4)

        if not (status := client_info.get('_html_ref')):
            status = 'An undefined login attempt'

        if not all([env.gmail_user, env.gmail_pass, env.recipient]) or status not in ['LOGIN SUCCESS',
                                                                                      'An undefined login attempt']:
            return

        if not (endpoint := ngrok.get_ngrok()):
            endpoint = env.endpoint or f"http://{':'.join(map(str, self.connection.getsockname()))}"

        with open(client_file, 'r') as client:
            attachment_info = client.read()

        body = f"A connection with status `{status}` has been made to your file server hosted at " \
               f"{endpoint}\n\nDetails of the client are below. If this was you, you may disregard this email. " \
               f"Otherwise stop the server immediately and rotate your credentials before restarting." \
               f"\n\n\n{attachment_info}\n\n"

        if env.host_dir == env.home_dir:
            base_path = PurePath(__file__)
            body += f"\n\n\nLogs: {endpoint}/" \
                    f"{(str(base_path).strip(base_path.name) + models.LogConfig.SERVER_LOG_FILE).strip(env.host_dir)}"

        response = SendEmail(gmail_user=env.gmail_user, gmail_pass=env.gmail_pass, recipient=env.recipient,
                             subject=f"WARNING: {status} was detected. {current_time}", attachment=client_file,
                             body=body, sender='FileWare').send_email()
        if response.ok:
            logger.info('Email has been sent successfully.')
        else:
            logger.error(response.body)

    def disable_cache(self) -> None:
        """Headers to force no-cache and site-data to expire."""
        if 'Cache-Control' in self.headers.keys():
            self.headers.replace_header(_name='Cache-Control', _value='no-cache, no-store, must-revalidate')
        else:
            self.headers.add_header(_name='Cache-Control', _value='no-cache, no-store, must-revalidate')
        self.headers.add_header(_name='Pragma', _value='no-cache')
        self.headers.add_header(_name='Expires', _value='0')
        self.headers.add_header(_name='Clear-Site-Data', _value='"cache", "cookies", "storage", "executionContexts"')


def _reset_auth() -> bool:
    """Tells if the authentication header has to be reset and cache to be cleared.

    See Also:
        Note that if an authentication is done at the end of 15 minutes, there will be a re-auth prompted.

    Returns:
        bool:
        True if it is the first login attempt, or it has been more than 15 minutes since the first/previous expiry.
    """
    if env.first_run:
        env.first_run = False
        return True
    elif time() - env.start_time > 900:
        env.start_time = time()
        return True



"""Place for all necessary arguments accessed across multiple modules."""

import os
import socket
from getpass import getuser
from time import time

import dotenv

if os.path.isfile('.env'):
    dotenv.load_dotenv(dotenv_path='.env', verbose=True, override=True)  # loads the .env file

start_time = time()  # set to the current time to reset the auth headers when timeout is reached
first_run = True  # set first_run to True to prompt first time auth regardless of stored cookies
authenticated = False  # set to False for the appropriate message to be written in the HTML file
renamed = []  # set to empty list to append dictionaries with re-namings done for index.html
notify = True  # notify is set to true by default unless the env vars are not present
endpoint = None  # endpoint is set to None to make it a global variable

STOPPER = False

host = socket.gethostbyname('localhost')
port = int(os.environ.get('port', 4443))

gmail_user = os.environ.get('gmail_user')
gmail_pass = os.environ.get('gmail_pass')
recipient = os.environ.get('recipient')
username = os.environ.get('username', os.environ.get('USER', getuser()))
password = os.environ.get('password', 'FileServer')
home_dir = os.path.expanduser('~')
host_dir = os.environ.get('host_path', home_dir)
ngrok_auth = os.environ.get('ngrok_auth')



from logging import INFO, basicConfig, getLogger
from os import environ, listdir, system
from pathlib import PurePath
from subprocess import check_output
from sys import exit


class Volume:
    """Initiates Volume object to mount, unmount and stop the disk usage.

    >>> Volume

    """

    def __init__(self, label):
        """Stores label, password and initiates logging.

        Args:
            label: Name of the volume connected.
        """
        basicConfig(format="%(asctime)s - [%(levelname)s] - %(name)s - %(funcName)s - Line: %(lineno)d - %(message)s",
                    level=INFO)
        self.logger = getLogger(PurePath(__file__).stem)  # gets the current file name
        self.volume_label = label
        self.password = environ.get('PASSWORD')
        self.mount_uuid = None

    def stop_usage(self) -> None:
        """Kills all the background processes utilizing the volume to stop the disk usage."""
        if self.volume_label not in listdir('/Volumes'):
            exit(f'{self.volume_label} is not connected to/mounted on your {device_model()}.')
        if not self.password:
            exit("Add 'PASSWORD={YOUR SYSTEM PASSWORD}' as env variable to stop disk usage.")
        pid_check = check_output(f"echo {self.password} | sudo -S lsof '/Volumes/{self.volume_label}' 2>&1;",
                                 shell=True)
        pid_list = pid_check.decode('utf-8').split('\n')
        self.logger.info(f'Number of processes using {self.volume_label}: {len(pid_list) - 1}')
        for id_ in pid_list:
            system(f'kill -9 {id_.split()[1]} >/dev/null 2>&1') if id_ else None
        self.logger.info(f'{len(pid_list) - 1} active processes have been terminated.')

    def unmount_disk(self) -> None:
        """Ejects the disk from parent device."""
        self.logger.info('Accessing stop_usage')
        self.stop_usage()
        disk_check = check_output("diskutil list 2>&1;", shell=True)
        disk_list = disk_check.decode('utf-8').split('\n\n')
        condition = '(external, physical):'
        for disk in disk_list:
            if disk and condition in disk:
                unmount_uuid = disk.split('\n')[0].strip(condition)
                disk_info = disk.split('\n')[-1]
                if self.volume_label in disk_info:
                    self.mount_uuid = unmount_uuid
                    system(f'diskutil unmountDisk {unmount_uuid} > /dev/null 2>&1;')
                    self.logger.info(f'Disk {unmount_uuid} with Name {self.volume_label} has been unmounted from your '
                                     f'{device_model()}')
                    break

    def mount_disk_by_uuid(self) -> None:
        """Tries to mount the disk using UUID which is updated by the `unmount_disk` function.

        Calls `mount_disk_by_label` if UUID is None.
        """
        if self.mount_uuid:
            system(f'diskutil mountDisk {self.mount_uuid} > /dev/null 2>&1;')  # mount using disk volume/UUID
        else:
            self.logger.info('Accessing mount_disk_by_label')
            self.mount_disk_by_label()

    def mount_disk_by_label(self) -> None:
        """Uses a terminal command to mount the disk using the volume name."""
        system(f'diskutil mount "{self.volume_label}" > /dev/null 2>&1;')  # mount using disk label
        if self.volume_label in listdir('/Volumes'):
            self.logger.info(f'Disk {self.volume_label} has been mounted on your {device_model()}')
        else:
            self.logger.info(f'{self.volume_label} is not connected or mounted on your {device_model()}.')


def extract_str(input_) -> str:
    """Extracts strings from the received input.

    Args:
        input_: Un-formatted string.

    Returns:
        str:
        A perfect string.

    """
    return ''.join([i for i in input_ if not i.isdigit() and i not in [',', '.', '?', '-', ';', '!', ':']])


def device_model() -> str:
    """Uses `sysctl` command to get the device model and version information.

    Returns:
        str:
        Model of the parent device.

    """
    device = (check_output("sysctl hw.model", shell=True)).decode('utf-8').split('\n')  # gets model info
    result = list(filter(None, device))[0]  # removes empty string ('\n')
    model = extract_str(result).replace('hwmodel', '').strip()
    return model



import functools
import os
import ssl
import warnings
from http.server import HTTPServer
from multiprocessing import Process
from socket import AF_INET, SOCK_DGRAM, socket
from typing import AnyStr, Union
from uuid import UUID

from . import cert, env, models, ngrok
from .server import Authenticator

logger = models.server_logger()


class FileWare:
    """Class to instantiate FileWare object, and run the file server in a thread.

    >>> FileWare

    """

    def __init__(self, port: Union[int, str] = None, username: AnyStr = None, password: AnyStr = None,
                 host_dir: AnyStr = None, ngrok_auth: Union[UUID, str] = None, gmail_user: str = None,
                 gmail_pass: str = None, recipient: str = None) -> None:
        """Initiates the ``FileWare``.

        Args:
            port: Port number in which the file server is running.
            username: Username to access fileserver.
            password: Password to access fileserver.
            host_dir: Takes the path to serve as an argument. Can also be loaded via env vars.
            ngrok_auth: Ngrok auth token for tunneling.
            gmail_user: Username for email notification.
            gmail_pass: Password for email notification.
            recipient: Email address to receive notification.

        See Also:
            - All these arguments can be loaded via env vars by placing the key-value pairs in a ``.env`` file.
            - The ``.env`` file should be stored in the current working directory.
        """
        if port:
            env.port = port
        if gmail_user:
            env.gmail_user = gmail_user
        if gmail_pass:
            env.gmail_pass = gmail_pass
        if recipient:
            env.recipient = recipient
        if username:
            env.username = username
        if password:
            env.password = password
        if ngrok_auth:
            env.ngrok_auth = ngrok_auth
        if host_dir:
            if os.path.isdir(host_dir):
                env.host_dir = host_dir
            else:
                warnings.warn(f"The specified path: {host_dir} doesn't exist. Defaulting to {env.host_dir}")

        self._ngrok_object = ngrok.Ngrok()
        self._public_url = self._ngrok_object.connect()
        self._ngrok_process = Process(target=self._ngrok_object.tunnel)
        ssh_dir = env.home_dir + os.path.sep + os.path.join('.ssh')
        self._cert_file = os.path.expanduser(ssh_dir) + os.path.sep + "cert.pem"
        self._key_file = os.path.expanduser(ssh_dir) + os.path.sep + "key.pem"

        if not self._public_url:
            self._host_on_ip()

        handler_class = functools.partial(
            Authenticator,
            username=env.username,
            password=env.password,
            directory=env.host_dir
        )
        self._http_server = HTTPServer(server_address=(env.host, env.port), RequestHandlerClass=handler_class)

    def _host_on_ip(self):
        """Changes hostname to local IP address and generates a self-signed certificate.

        See Also:
            - Hosts on ``localhost`` (127.0.0.1) if Ngrok connection is successful, otherwise hosts on local IP address.
            - Ngrok uses its own certificate, so the self-signed cert is generated only when hosting on local IP.
        """
        ip_socket = socket(AF_INET, SOCK_DGRAM)
        ip_socket.connect(("8.8.8.8", 80))
        env.host = ip_socket.getsockname()[0]
        if not all([os.path.isfile(self._cert_file), os.path.isfile(self._key_file)]):
            cert.generate_cert(common_name='*.ngrok.com', cert_file=self._cert_file, key_file=self._key_file)

    def initiate_connection(self) -> str:
        """Initiates fileserver connection after trying to trigger ngrok tunnel.

        See Also:
            - Checks for ``cert.pem`` and ``key.pem`` files in ~home/ssh path.
            - If not generates a self-signed certificate using ``OpenSSL``
            - If ngrok tunnel is running on the port already, initiates file server on localhost else uses local IP.
        """
        if not self._public_url and all([os.path.isfile(self._cert_file), os.path.isfile(self._key_file)]):
            self._http_server.socket = ssl.wrap_socket(sock=self._http_server.socket, server_side=True,
                                                       certfile=self._cert_file, keyfile=self._key_file)
            endpoint = f"https://{':'.join(map(str, self._http_server.server_address))}"
        else:
            endpoint = f"http://{':'.join(map(str, self._http_server.server_address))}"

        logger.info(f"Initiating file server at: {endpoint}")
        if self._public_url:
            logger.info(f"Hosted at public endpoint: {self._public_url}")

        if self._public_url:
            return self._public_url
        else:
            return self._public_url

    def shutdown(self) -> None:
        """Terminates the http server, ngrok process and the socket server."""
        env.STOPPER = True
        if self._ngrok_process.is_alive():
            self._ngrok_process.join(timeout=1e-2)
            self._ngrok_object.stop_tunnel()
        self._http_server.server_close()
        self._http_server.shutdown()

    def serve(self) -> None:
        """Starts the file server session and ngrok tunnel (if available)."""
        if self._public_url:
            self._ngrok_process.start()
        try:
            self._http_server.serve_forever()
        except KeyboardInterrupt:
            [(os.rename(new_name, old_name), logger.critical(f'Reverted {new_name} to {old_name}'))
             for each in env.renamed for old_name, new_name in each.items() if env.renamed]
            self.shutdown()



import logging.config
import os
from datetime import datetime
from pathlib import Path

os.makedirs('logs') if not os.path.isdir('logs') else None


def server_logger():
    """Creates a log file for server logs and configures the logging config for server logger.

    Returns:
        logging.Logger:
        Returns logger config for server.
    """
    if not os.path.isfile(LogConfig.SERVER_LOG_FILE):
        Path(LogConfig.SERVER_LOG_FILE).touch()

    logging.config.dictConfig(config=LogConfig.LOGGING_CONFIG)
    return logging.getLogger('server')


def ngrok_logger() -> logging.Logger:
    """Creates a log file for ngrok logs and configures the logging config for ngrok logger.

    Returns:
        logging.Logger:
        Returns logger config for ngrok.
    """
    if not os.path.isfile(LogConfig.TUNNEL_LOG_FILE):
        Path(LogConfig.TUNNEL_LOG_FILE).touch()

    logging.config.dictConfig(config=LogConfig.LOGGING_CONFIG)
    return logging.getLogger('tunnel')


class LogConfig:
    """Custom log configuration.

    >>> LogConfig

    """

    LOG_LEVEL = "DEBUG"

    SERVER_LOG_FILE = datetime.now().strftime('logs' + os.path.sep + 'fileserver_%m_%d_%Y.log')
    TUNNEL_LOG_FILE = datetime.now().strftime('logs' + os.path.sep + 'tunnel_%m_%d_%Y.log')

    FILE_LOG_FORMAT = '%(asctime)s - %(levelname)s - [%(module)s:%(lineno)d] - %(funcName)s - %(message)s'

    LOGGING_CONFIG = {
        "version": 1,
        "disable_existing_loggers": True,
        "formatters": {
            "server": {
                "format": FILE_LOG_FORMAT,
                "filename": SERVER_LOG_FILE,
                "datefmt": "%b %d, %Y %H:%M:%S"
            },
            "tunnel": {
                "format": FILE_LOG_FORMAT,
                "filename": TUNNEL_LOG_FILE,
                "datefmt": "%b %d, %Y %H:%M:%S"
            }
        },
        "handlers": {
            "server": {
                "formatter": "server",
                "class": "logging.FileHandler",
                "filename": SERVER_LOG_FILE,
            },
            "tunnel": {
                "formatter": "tunnel",
                "class": "logging.FileHandler",
                "filename": TUNNEL_LOG_FILE,
            }
        },
        "loggers": {
            "server": {"handlers": ["server"], "level": LOG_LEVEL, "filename": SERVER_LOG_FILE},
            "tunnel": {"handlers": ["tunnel"], "level": LOG_LEVEL, "filename": TUNNEL_LOG_FILE}
        }
    }



import os

from setuptools import setup

from version import version_info

classifiers = [
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Information Technology',
    'Operating System :: OS Independent',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.9',
    'Topic :: Internet :: WWW/HTTP :: HTTP Servers'
]


def read(name: str) -> str:
    """https://pythonhosted.org/an_example_pypi_project/setuptools.html#setting-up-setup-py - reference."""
    return open(os.path.join(os.path.dirname(__file__), name)).read()


setup(
    name='fileware',
    version='.'.join(str(c) for c in version_info),
    description='Python module to, serve files in physical memory to localhost and tunnel to a public endpoint.',
    long_description=read('README.md') + '\n\n' + read('CHANGELOG'),
    long_description_content_type='text/markdown',
    url='https://github.com/thevickypedia/fileware',
    author='Vignesh Sivanandha Rao',
    author_email='svignesh1793@gmail.com',
    License='MIT',
    classifiers=classifiers,
    keywords='personal-cloud, file-server, ssl',
    packages=['.fileware'],
    python_requires=">=3.8",
    include_package_data=True,
    install_requires=read(name='requirements.txt').splitlines(),
    project_urls={
        'Docs': 'https://thevickypedia.github.io/fileware',
        'Bug Tracker': 'https://github.com/thevickypedia/fileware/issues'
    }
)



import binascii
import json
import os
import time
from urllib.error import HTTPError
from urllib.request import urlopen

import OpenSSL

from . import env


def ip_info() -> dict:
    """Gets public IP information.

    Returns:
        dict:
        JSON loaded information from public IP information.
    """
    try:
        return json.load(urlopen('https://ipinfo.io/json'))
    except HTTPError:
        pass
    try:
        return json.load(urlopen('http://ip.jsontest.com'))
    except HTTPError:
        pass
    return {}


def _get_serial() -> bytes:
    """Generates a serial number for the self-signed SSL.

    See Also:
        - This function is not called, but it is here only as a just in case measure to insert serial number manually.
        - Serial Number is a unique identifier assigned by the CA which issued the certificate.

    Returns:
        bytes:
        Encoded serial number for the certificate.
    """
    serial_hex = binascii.hexlify(os.urandom(18)).decode().upper()
    return " ".join(serial_hex[i:i + 2] for i in range(0, len(serial_hex), 2)).encode('UTF-8')


def _generate_serial_hash(byte_size: int = 18, int_size: int = 36) -> int:
    """Generates a hashed serial number.

    Args:
        byte_size: Size of the bytes object containing random bytes.
        int_size: Size of the base int.

    Returns:
        int:
        Returns the hashed serial.
    """
    return int(binascii.hexlify(os.urandom(byte_size)).decode().upper(), int_size)


def generate_cert(common_name: str,
                  email_address: str = None,
                  country_name: str = None,
                  locality_name: str = None,
                  state_or_province_name: str = None,
                  organization_name: str = None,
                  organization_unit_name: str = "Information Technology",
                  validity_start_in_seconds: int = 0,
                  validity_end_in_seconds: int = 10 * 365 * 24 * 60 * 60,
                  key_file: str = "key.pem",
                  cert_file: str = "cert.pem",
                  key_size: int = 4096) -> bool:
    """Creates a private/self-signed certificate.

    Args:
        common_name: DNS name of the origin.
        email_address: Contact email address for the cert. Defaults to ``USER@expose-localhost.com``
        country_name: Name of the country. Defaults to ``US``
        locality_name: Name of the city. Defaults to ``New York``
        state_or_province_name: Name of the state/province. Defaults to ``New York``
        organization_name: Name of the organization. Defaults to ``common_name``
        organization_unit_name: Name of the organization unit. Defaults to ``Information Technology``
        validity_start_in_seconds: From when the cert validity begins. Defaults to ``0``.
        validity_end_in_seconds: Expiration duration of the cert. Defaults to ``10 years``
        key_file: Name of the key file.
        cert_file: Name of the certificate.
        key_size: Size of the public key. Defaults to 4096.

    See Also:
        Use ``openssl x509 -inform pem -in cert.crt -noout -text`` to look at the generated cert using openssl.

    Returns:
        bool:
        Boolean flag to indicate whether ``cert.pem`` and ``key.pem`` files are empty.
    """
    ip_ = ip_info()
    country_name = country_name or ip_.get('country', 'US')
    locality_name = locality_name or ip_.get('city', 'New York')
    state_or_province_name = state_or_province_name or ip_.get('region', 'New York')

    if key_size not in [2048, 4096]:
        raise ValueError('Certificate key size should be either 2048 or 4096.')
    signature_bytes = 256 if key_size == 2048 else 512  # Refer: https://crypto.stackexchange.com/a/3508

    # Creates a key pair
    key = OpenSSL.crypto.PKey()
    key.generate_key(type=OpenSSL.crypto.TYPE_RSA, bits=key_size)

    # Creates a self-signed cert
    cert = OpenSSL.crypto.X509()
    cert.get_subject().C = country_name
    cert.get_subject().ST = state_or_province_name
    cert.get_subject().L = locality_name
    cert.get_subject().O = organization_name or common_name[0].upper() + common_name.partition('.')[0][1:]  # noqa: E741
    cert.get_subject().OU = organization_unit_name
    cert.get_subject().CN = common_name
    cert.get_subject().emailAddress = email_address or f"{env.username}@fileware.com"
    cert.set_serial_number(serial=cert.get_serial_number() or _generate_serial_hash())
    cert.gmtime_adj_notBefore(amount=validity_start_in_seconds)
    cert.gmtime_adj_notAfter(amount=validity_end_in_seconds)
    cert.set_issuer(issuer=cert.get_subject())
    cert.set_pubkey(pkey=key)
    # noinspection PyTypeChecker
    cert.sign(pkey=key, digest=f'sha{signature_bytes}')

    # Writes the cert file into specified names
    with open(cert_file, "w") as f:
        f.write(OpenSSL.crypto.dump_certificate(type=OpenSSL.crypto.FILETYPE_PEM, cert=cert).decode("utf-8"))
    with open(key_file, "w") as f:
        f.write(OpenSSL.crypto.dump_privatekey(type=OpenSSL.crypto.FILETYPE_PEM, pkey=key).decode("utf-8"))

    if os.stat(cert_file).st_size != 0 and os.stat(key_file).st_size != 0:
        time.sleep(1)
        return True
